# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""🔲 License Plate Detection model."""

from typing import Any, Dict

import numpy as np

from peekingduck.pipeline.nodes.model.yolov4_license_plate import lp_detector_model
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Initializes and uses YOLO model to infer bboxes from image frame.

    This customized YOLO node is capable of detecting objects from a single
    class (License Plate). It uses YOLOv4 by default and can be changed to
    use YOLOv4-tiny if FPS is critical over accuracy.

    Inputs:
        |img_data|

    Outputs:
        |bboxes_data|

        |bbox_labels_data|

        |bbox_scores_data|

    Configs:
        model_type (:obj:`str`): **{"v4", "v4tiny"}, default="v4"**. |br|
            Defines the type of YOLO model to be used.
        weights_parent_dir (:obj:`Optional[str]`): **default = null**. |br|
            Change the parent directory where weights will be stored by
            replacing ``null`` with an absolute path to the desired directory.
        iou_threshold (:obj:`float`): **[0, 1], default = 0.3**. |br|
            Overlapping bounding boxes above the specified IoU (Intersection
            over Union) threshold are discarded.
        score_threshold (:obj:`float`): **[0, 1], default = 0.1**. |br|
            Bounding box with confidence score less than the specified
            confidence score threshold is discarded.

    References:
        YOLOv4: Optimal Speed and Accuracy of Object Detection:
        https://arxiv.org/pdf/2004.10934v1.pdf

        Model weights trained using pretrained weights from Darknet:
        https://github.com/AlexeyAB/darknet

    .. versionchanged:: 1.2.0
        ``yolo_iou_threshold`` is renamed to ``iou_threshold``.

    .. versionchanged:: 1.2.0
        ``yolo_score_threshold`` is renamed to ``score_threshold``.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.model = lp_detector_model.Yolov4(self.config)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads the image input and returns the bboxes of the specified
        objects chosen to be detected.

        Args:
            inputs (dict): Dictionary of inputs with key "img".
        Returns:
            outputs (dict): bbox output in dictionary format with keys
            "bboxes", "bbox_labels", and "bbox_scores".
        """
        bboxes, labels, scores = self.model.predict(inputs["img"])
        bboxes = np.clip(bboxes, 0, 1)

        outputs = {
            "bboxes": bboxes,
            "bbox_labels": labels,
            "bbox_scores": scores,
        }

        return outputs



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Postprocessing functions for HRNet
"""

from typing import List, Tuple
import numpy as np

# fmt: off
SKELETON = [
    [16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13], [6, 7],
    [6, 8], [7, 9], [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4], [3, 5],
    [4, 6], [5, 7],
]
# fmt: on


def scale_transform(
    keypoints: np.ndarray, in_scale: List[int], out_scale: List[int]
) -> np.ndarray:
    """Transform points from input scale to out scale.

    Args:
        keypoints (np.array): array of detected keypoints
        in_scale (list): the input scale of xy
        out_scale (list): the output scale of xy

    Returns:
        the scaled points from input scale to output scale
    """
    in_to_out_scale = np.array(out_scale) / np.array(in_scale)
    keypoints = keypoints * in_to_out_scale
    return keypoints


def affine_transform_xy(
    keypoints: np.ndarray, affine_matrices: np.ndarray
) -> np.ndarray:
    """Apply respective affine transform on array of points.

    Args:
        keypoints (np.array): array sequence of points (x, y)
        affine_matrices (np.array) - array of 2x3 affine transform matrix

    Returns:
        array of transformed points
    """
    transformed_matrices = []
    keypoints = np.dstack(
        (keypoints, np.ones((keypoints.shape[0], keypoints.shape[1], 1)))
    )
    for affine_matrix, keypoint in zip(affine_matrices, keypoints):
        transformed_keypoint = np.dot(affine_matrix, keypoint.T)
        transformed_matrices.append(transformed_keypoint.T)

    return np.array(transformed_matrices)


def reshape_heatmaps(heatmaps: np.ndarray) -> np.ndarray:
    """Helper function to reshape heatmaps to required shape

    Args:
        heatmaps (np.ndarray): Outputs heatmaps from hrnet network

    Returns:
        np.ndarray: reshaped heatmaps
    """
    batch, _, _, num_joints = heatmaps.shape

    heatmaps_reshaped = np.transpose(heatmaps, axes=(0, 3, 1, 2))
    heatmaps_reshaped = heatmaps_reshaped.reshape((batch, num_joints, -1))

    return heatmaps_reshaped


def get_valid_keypoints(
    keypoints: np.ndarray, keypoint_scores: np.ndarray, batch: int, min_score: int
) -> Tuple[np.ndarray, np.ndarray]:
    """Helper function to get visible keypoints

    Args:
        keypoints (np.ndarray): array of detected keypoints
        keypoint_scores (np.ndarray): array of keypoint scores
        batch (int): number of detected bboxes
        min_score (int): min score threshold

    Returns:
        Tuple[np.ndarray, np.ndarray]: array of keypoints above threshold and keypoint mask
    """
    score_masks = keypoint_scores > min_score
    kp_masks = (
        np.repeat(keypoint_scores > 0.0, 2).reshape(batch, 17, -1).astype(np.float32)
    )
    keypoints *= kp_masks
    return keypoints, score_masks


def get_keypoint_conns(rel_keypoints: np.ndarray, masks: np.ndarray) -> np.ndarray:
    """Helper function to get keypoint connections

    Args:
        rel_keypoints (np.ndarray): Array of normalized keypoints
        masks (np.ndarray): Array of keypoint masks

    Returns:
        np.ndarray: Array of keypoint connections
    """
    compiled_connections = []
    for keypoint, mask in zip(rel_keypoints, masks):
        single_conn = _get_keypoint_of_single_pose(keypoint, mask)
        compiled_connections.append(single_conn)
    return np.array(compiled_connections, dtype=object)


def _get_keypoint_of_single_pose(keypoint: np.ndarray, mask: np.ndarray) -> np.ndarray:
    connections = []
    for start_joint, end_joint in SKELETON:
        if mask[start_joint - 1] and mask[end_joint - 1]:
            connections.append((keypoint[start_joint - 1], keypoint[end_joint - 1]))
    return np.array(connections)



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Preprocessing functions for HRNet
"""

from typing import Tuple

import cv2
import numpy as np


def project_bbox(bboxes: np.ndarray, size: Tuple[int, int]) -> np.ndarray:
    """Project the normalized bbox to specified image coordinates.

    Args:
        bboxes (np.ndarray): Array of detected bboxes in (top left, btm right)
            format.
        size (Tuple[int, int]): Width and height of image space.

    Returns:
        (np.ndarray): Array of bboxes in (x, y, w, h) format where x, y is top
            left coordinate.
    """
    width, height = size[0] - 1, size[1] - 1

    bboxes[:, 0] = np.clip(bboxes[:, 0], 0, 1)
    bboxes[:, 1] = np.clip(bboxes[:, 1], 0, 1)
    bboxes[:, 2] = np.clip(bboxes[:, 2], bboxes[:, 0], 1)
    bboxes[:, 3] = np.clip(bboxes[:, 3], bboxes[:, 1], 1)

    bboxes[:, [0, 2]] *= width
    bboxes[:, [1, 3]] *= height

    bboxes[:, 2] = bboxes[:, 2] - bboxes[:, 0]
    bboxes[:, 3] = bboxes[:, 3] - bboxes[:, 1]

    return bboxes


def box2cs(bboxes: np.ndarray, aspect_ratio: float) -> np.ndarray:
    """Converts bounding box defined by top left (x, y, w, h) to its
    center (x, y, w, h). The bounding box is also expanded to meet the input
    aspect ratio.

    Args:
        bboxes (np.ndarray): Array of bboxes in the form of top left
            (x, y, w, h).
        aspect_ratio (float): W:H ratio for the new box.

    Returns:
        (np.ndarray): Array of bboxes in center (x, y, w, h) format that meets
        the aspect ratio.
    """
    bboxes[:, 0] = bboxes[:, 0] + bboxes[:, 2] * 0.5
    bboxes[:, 1] = bboxes[:, 1] + bboxes[:, 3] * 0.5

    req_w = aspect_ratio * bboxes[:, 3]
    bboxes[:, 3][bboxes[:, 2] > req_w] = (
        bboxes[bboxes[:, 2] > req_w][:, 2] / aspect_ratio
    )
    bboxes[:, 2][bboxes[:, 2] < req_w] = (
        bboxes[bboxes[:, 2] < req_w][:, 3] * aspect_ratio
    )

    return bboxes


def crop_and_resize(
    frame: np.ndarray, bboxes: np.ndarray, out_size: Tuple[int, int]
) -> Tuple[np.ndarray, np.ndarray]:
    """Crop a region from frame specified by its center and size. The
    cropped region is resized to out_size.

    Args:
        frame (np.ndarray): Image in numpy array.
        bboxes (np.ndarray): Bboxes center (x, y, w, h) coordinates.
        out_size (tuple): Cropped region will be resized to out_size.

    Returns:
        (Tuple[np.ndarray, np.ndarray]): The resized and cropped region array
        and the affine transform matrix to map a point in cropped image
        coordinate space to source frame coordinate space.
    """
    translate_x = bboxes[:, 0] - (bboxes[:, 2] - 1) * 0.5
    translate_y = bboxes[:, 1] - (bboxes[:, 3] - 1) * 0.5
    scale_x = bboxes[:, 2] / out_size[0]
    scale_y = bboxes[:, 3] / out_size[1]
    zero_mat = np.zeros((len(translate_x),))

    x_mat = np.column_stack((scale_x, zero_mat, translate_x))
    y_mat = np.column_stack((zero_mat, scale_y, translate_y))
    affine_matrices = np.concatenate((x_mat, y_mat), axis=1)
    affine_matrices = affine_matrices.reshape((-1, 2, 3))

    transformed_images = [
        cv2.warpAffine(
            frame, x, out_size, flags=cv2.INTER_LINEAR | cv2.WARP_INVERSE_MAP
        )
        for x in affine_matrices
    ]
    return transformed_images, affine_matrices
